/*
  Copyright 2007-2020 Guillaume Hanrot, Vincent Lefèvre, Damien Stehlé and
  Paul Zimmermann.  

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

*/

#define VERSION "4.0"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <omp.h>
#include <gmp.h>
#include <mpfr.h>
#include "mat.h"
#include "io.h"
#include "poly.h"
#include "hensel.h" /* must come before slz.h */
#include "slz.h"
#include "taylor.h"
#include "bacsel.h"
#include "heuristic.h"
#include "function.h"
#include "utils.h"
#include "getexp.h"
#include "delta_eta.h"

// #define VBSE
// #define VBSE1
// #define DEB
// #define DEBDEB
// #define DEB2
// #define VERIF
// #define AUTOMATIC
/*#define NO_OVERPASS_FAILURES
#define EVAL_TAYLOR 
#define NBIT_TAYLOR 1000
#define EVAL_CHECK
#define NBIT_CHECK 1000 
*/
#ifdef AUTOMATIC
#ifndef FAILURE_LOWERBOUND
#define FAILURE_LOWERBOUND 1.0
#endif
#ifndef FAILURE_UPPERBOUND
#define FAILURE_UPPERBOUND 4.0
#endif
#endif

/* Choose your basis. Default: 10. */
#ifndef BASIS
#define BASIS 10
#endif
  
#define LARGE_PREC 1200 /* FIXME: set according to n */

#define CHECK_HENSEL 

/* global variables that do not depend on [t0,t1] */
mpfr_t tau; /* elementary step */
#ifdef SINBIG
mpfr_t pi, twopi; /* Pi and 2*Pi */
#endif
mpfr_t one_half; /* 1/2 */
int near; /* rounding mode */
long Failures = 0, Good = 0, Number = 0;
mpz_t AverageT;

int verbose = 1;

/********************************************************************/
/* checks that the retuned possible solutions in  Sol are indeed 
   solutions, in which case they are printed */
/********************************************************************/

/* For all x0 \in Res, print x iff. | NN. f( (qx0+tm)/N ) cmod 1| \leq 1/M */
/* and qx0+tm < t1 */

int
Check (mpz_t *Res, int Nb, mpz_t tm, mpz_t t1, mpfr_t N, mpfr_t NN, 
       mpfr_t oneoverM, mpz_t T, int e_in, int n, mpz_t q MAYBE_UNUSED,
       mpfr_t yy MAYBE_UNUSED)
{
  int i, k = 0;
  mpfr_t y1, y2;

  mpfr_init (y1);
  mpfr_init (y2);

  for (i = 0; i < Nb; i++)  {
    if (mpz_cmpabs(Res[i], T) <= 0) { /* | Res[i] | <= T  */
#ifdef SINBIG
      mpz_mul(Res[i], Res[i], q); 
#endif
      mpz_add(Res[i], Res[i], tm);

      if (mpz_cmp (Res[i], t1) < 0){
	mpfr_set_z (y1, Res[i], MPFR_RNDN);
	mpfr_div (y1, y1, N, MPFR_RNDN); /* (Res[i]+tm)/N */
	
	EVAL(y1, MPFR_RNDN); 

	mpfr_mul (y1, y1, NN, MPFR_RNDN); /* NN * f((tm + Res[i]) / N) */
	
	if (near == 1) mpfr_sub (y1, y1, one_half, MPFR_RNDN);
	
	mpfr_round (y2, y1);
	mpfr_sub (y1, y1, y2, MPFR_RNDN);
	mpfr_abs (y1, y1, MPFR_RNDN); /* { NN * f((tm + Res[i]) / N) } */
	
#if DEBUG >= 2
	printf ("oneover M is ");
	mpfr_out_str (stdout, 10, 10, oneoverM, MPFR_RNDN);
	printf ("\ny1 is ");
	mpfr_out_str (stdout, 10, 10, y1, MPFR_RNDN);
	printf ("\n");
#endif

	if ((mpfr_cmp (y1, oneoverM) <= 0)) /* Indeed a solution. */
#pragma omp critical
          {
            k ++;
            printf ("*** x=");
            mpz_out_str (stdout, 10, Res[i]);
#if BASIS==10
            printf("E%d, distance is: ", e_in - n);
#else
            printf("*%d^%d, distance is: ", BASIS, e_in - n);
#endif
            /* if mode is 'all', since we multiplied NN by 2,
               we divide the distance by 2 to get comparable distances */
            if (near == 2) mpfr_div_2ui (y1, y1, 1, MPFR_RNDN);
            mpfr_out_str (stdout, 10, 10, y1, MPFR_RNDN);
            printf ("\n");
            fflush (stdout);
          }
      }
    }
  }

  mpfr_clear (y1);
  mpfr_clear (y2);

  return k;
}

int
remove_doubles(mpz_t *Res, int Nb)
{
  int i, j, k; 

  /* Remove several occurrences of the same number */ 
  for (i = 0; i < Nb; i++)
    for (j = i+1; j < Nb; j++){
      if (mpz_cmp (Res[i], Res[j]) == 0){
	for (k = j; k < Nb-1; k++)
	  mpz_set (Res[k], Res[k+1]);
	Nb--;
      }
    }

  return Nb; 
}
/**************************************************************************/

/* process the interval [t0,t1-1] (t1 is excluded) */
static void
doit (mpz_t t0, mpz_t t1, double n, double nn, double m, double t, int d,
      int alpha, int e_in, mpz_t q)
{
  int Nb, success, i, number = 0, npols, ncoefs, *ok, j, nb;
  mpfr_prec_t e_out, e_out2;
  double total_T = 0.0;
  mpz_t *Res = NULL, C, initial_C, initial_T, T, tm, ztmp, ztmp2, Calpha, NZ;
  mpz_t *PP, *Ptmp, t0current, done;
  mpfr_t tmp, tmp2, *P, M, oneoverM, Mprime, NN, Nin, initial_eps, eps;
  long failures = 0, good = 0; /* number of failures */
  mat_ZZ W;
  unsigned long int time_4_lll=0, time_4_Hensel=0;
  /* LLL variables */
  mpz_t         **B_lll;
  mpfr_t yy; /* auxiliary variable used in EVAL() and TAYLOR() */
  mpz_t *Sol[2];
  hensel_t h;

  ncoefs = ((alpha + 1) * (d * alpha + 2)) / 2;
  npols  = ((alpha + 1) * (alpha + 2)) / 2;   /* Nb. columns */
  nb = d * alpha * d * alpha; /* Bezout */
#ifdef CHECKHENSEL
  pari_init(100000000, 100000); 
#endif

  mpz_init (C); mpz_init (initial_C); mpz_init (initial_T); mpz_init (T); 
  mpz_init (tm); mpz_init (ztmp); mpz_init (ztmp2); mpz_init (Calpha); 
  mpz_init (NZ);

  Sol[0] = (mpz_t *) malloc (nb * sizeof(mpz_t));
  Sol[1] = (mpz_t *) malloc (nb * sizeof(mpz_t));    
  for (j = 0; j < nb; j++)
    {
      mpz_init (Sol[0][j]); 
      mpz_init (Sol[1][j]);
    }

  mpfr_init (Nin);
  mpfr_init (tmp);
  mpfr_init (tmp2);

  mpfr_init2 (NN, 53);
  mpfr_init2 (eps, 53); 
  mpfr_init2 (M, 53);
  mpfr_init2 (Mprime, 53); 
  mpfr_init2 (oneoverM, 53); 
  mpfr_init2 (initial_eps, 53);
  mpfr_init2 (yy, LARGE_PREC); 
  

  ok   = (int *) malloc ((npols+1) * sizeof (int));
  PP   = init_vec (d+1);
  Ptmp = init_vec (ncoefs);

  P = (mpfr_t *) malloc ((d+1)*sizeof (mpfr_t));
  for (i = 0; i<=d; i++) mpfr_init (P[i]);

  init_matrix (&W, npols, ncoefs);

  allocate_space_heuristic (&B_lll, npols, ncoefs);

  /***  END OF Initialization and argument parsing ***/ 

  /***  BEGINNING OF computing auxiliary variables ***/ 
  mpfr_set_d (tmp, n, MPFR_RNDN);
  mpfr_ui_pow (Nin, BASIS, tmp, MPFR_RNDN); /* B^p \pm 0.5 ulp(B^p) */ 
  mpfr_get_z (NZ, Nin, MPFR_RNDN); /* same, but mpz_t */ 
  check_t0_t1 (t0, t1, NZ);

  /* N = basis^n, NN=basis^nn, M=basis^m, T=initial_T=basis^t */
  // ERR. ANA.: exact value PREC >= (log_2 n) +1
  // ERR. ANA.: |N-BASIS^tmp| <= 0.5 ulp(BASIS^n).

  mpz_ui_pow_ui (T, BASIS, abs(e_in)); /* T = B^e_{in} */

  if (e_in > 0) 
    mpfr_div_z (Nin, Nin, T, MPFR_RNDN); 
  else 
    mpfr_mul_z (Nin, Nin, T, MPFR_RNDN); /* B^{p - e_{in}} */

  mpfr_set_d (tmp, m, MPFR_RNDN);
  // ERR. ANA.: exact value
  mpfr_ui_pow (M, BASIS, tmp, MPFR_RNDN);
  // ERR. ANA.: |M-BASIS^m| <= 0.5 ulp (BASIS^m)
  mpfr_ui_div (oneoverM, 1, M, MPFR_RNDU);
  // ERR. ANA.: |oneoverM-BASIS^-m| <= 0.5 ulp (1/M)

  mpfr_set_d (tmp, t, MPFR_RNDN);
  mpfr_ui_pow (tmp, BASIS, tmp, MPFR_RNDN);
  mpfr_get_z (T, tmp, MPFR_RNDN);
  mpz_set (initial_T, T);

#if DEBUG >= 2
  printf ("Nin= "); mpfr_out_str (stdout, 10, 10, Nin, MPFR_RNDN);
  printf ("\nM= "); mpfr_out_str (stdout, 10, 10, M, MPFR_RNDN); printf ("\n"); 
#endif


  bnddiff (eps, initial_eps, tau, tmp, Nin, T, t0, t1, d); 


#ifndef SINBIG

  /* check that t0 and t1 have the same exponent */
  e_out = getexp (t0, e_in - n);
  mpz_sub_ui (ztmp, t1, 1);
  e_out2 = getexp (ztmp, e_in - n);
  if (e_out != e_out2)
    {
      fprintf (stderr, "Error, f(t0) and f(t1-1) have different exponents\n");
      exit (1);
    }

  mpfr_set_d (tmp, nn - (double) e_out, MPFR_RNDN);
  // ERR. ANA.: exact value.
  mpfr_ui_pow (NN, BASIS, tmp, MPFR_RNDN);
  // ERR. ANA.: |NN-BASIS^(nn-e_out)| <= 0.5 ulp(BASIS^(nn-e_out)).
  if (near == 2) mpfr_mul_ui (NN, NN, 2, MPFR_RNDN);
  // ERR. ANA.: |NN-2.BASIS^(nn-e_out)| <= 0.5 ulp(2*lastNN)
  //                                    <= ulp((1+0.5ulp(1))BASIS^(nn-e_out))
  //                                    <= 2 ulp(BASIS^(nn-e_out))

#if DEBUG >= 2
  printf ("\nNN= "); mpfr_out_str (stdout, 10, 10, NN, MPFR_RNDN);
#endif

  mpfr_mul (eps, eps, NN, MPFR_RNDU);
  mpfr_set (initial_eps, eps, MPFR_RNDN);


  /* Mprime := E(1/2[1/M+eps]) 
     The 1/2 factor takes into account numerical errors on the Taylor 
     coefficients.
  */

  mpfr_ui_div (Mprime, 1, M, MPFR_RNDN);
  mpfr_add (Mprime, Mprime, eps, MPFR_RNDN);
  mpfr_mul_2ui (Mprime, Mprime, 1, MPFR_RNDN); 
  mpfr_ui_div (Mprime, 1, Mprime, MPFR_RNDN);

  mpfr_get_z (C, Mprime, MPFR_RNDD);
  mpz_set (initial_C, C);
  mpz_pow_ui (Calpha, C, alpha);
#endif /* ifndef SINBIG */
  /*** END OF computing auxiliary variables ***/ 

  /* BEGINNING OF main loop. t0 starts as its initial value, and is 
     progressively increased, taking into account the part already
     studied. The loop ends when t0 reaches t1. At each step we have
     done all values < t0. */

  mpz_init_set (t0current, t0);
  mpz_init_set_ui (done, 0);

  hensel_init (h, d, alpha);

  while (mpz_cmp (t0current, t1) < 0) {

      mpz_mul (ztmp, q, T);
      mpz_add (tm, t0current, ztmp); /* tm is the middle */
      mpz_add (ztmp, tm, ztmp);      /* ztmp is the right bound */

#ifdef SINBIG
      /* Check that t0current and t1current have the same e_out */
      /* If it is not the case, halve T */
      e_out = getexp (t0current, e_in - n);

      mpfr_set_prec (tmp, 53);
      mpfr_set_d (tmp, nn - (double) e_out, MPFR_RNDN);
      // ERR. ANA.: exact value.
      mpfr_ui_pow (NN, BASIS, tmp, MPFR_RNDN);
      // ERR. ANA.: |NN-BASIS^(nn-e_out)| <= 0.5 ulp(BASIS^(nn-e_out)).
      if (near == 2) mpfr_mul_ui (NN, NN, 2, MPFR_RNDN);
      // ERR. ANA.: |NN-2.BASIS^(nn-e_out)| <= 0.5 ulp(2*lastNN)
      //                         <= ulp((1+0.5ulp(1))BASIS^(nn-e_out))
      //                         <= 2 ulp(BASIS^(nn-e_out))
      
      mpfr_set_prec (tmp, mpfr_get_default_prec ());

      e_out2 = getexp (ztmp, e_in - n);
      if (e_out != e_out2)
        goto halve_T;

#if DEBUG >= 2
      printf ("NN= "); mpfr_out_str (stdout, 10, 10, NN, MPFR_RNDN);
      printf ("\n");
#endif
      
      mpfr_mul (eps, initial_eps, NN, MPFR_RNDU);
      
      /* Mprime := E(1/2[1/M+eps]) 
	 The 1/2 factor takes into account numerical errors on the Taylor 
	 coefficients.
      */

      mpfr_ui_div (Mprime, 1, M, MPFR_RNDN);
      mpfr_add (Mprime, Mprime, eps, MPFR_RNDN);
      mpfr_mul_2ui (Mprime, Mprime, 1, MPFR_RNDN); 
      mpfr_ui_div (Mprime, 1, Mprime, MPFR_RNDN);

      mpfr_get_z (C, Mprime, MPFR_RNDD);
      if (mpz_sgn (C) == 0)
        goto halve_T;
      mpz_pow_ui (Calpha, C, alpha);
#endif /* SINBIG */
      
#if DEBUG >= 2
      printf ("t0current= "); mpz_out_str (stdout, 10, t0current); 
      printf ("\n"); 
      printf ("T="); mpz_out_str (stdout, 10, T); printf ("\n");
      printf ("tm="); mpz_out_str (stdout, 10, tm); printf ("\n");
#endif
      
#ifdef EVAL_TAYLOR
      int st_taylor = cputime(), cptr_taylor; 
      for (cptr_taylor = 0; cptr_taylor < NBIT_TAYLOR; cptr_taylor++){
#endif
	/* Compute the Taylor polynomial */
#ifdef SIN_OR_COS
	TAYLOR(tmp, tmp2, tm, Nin, P, NN, d);
#else
	TAYLOR(tmp, tm, Nin, P, NN, d);
#endif	
	if (near == 1) mpfr_sub (P[0], P[0], one_half, MPFR_RNDN);
#ifdef EVAL_TAYLOR
      }
      printf("Taylor %d\n", (cputime() - st_taylor)/NBIT_TAYLOR); 
#endif
      
#if DEBUG >= 2
      printf ("P:="); print_uni_pol_mpfr(P, d); printf("\n");
#endif
      
      /* Look for the solutions of the equation |P(x_0/T) cmod 1| < 1/C, 
	 where  |x0| \leq T and C = floor (1/(1/M+eps)).*/
      success = slz (&Res, &Nb, P, C, T, d, alpha, tmp, &W, 
		     npols, ncoefs, ztmp, Calpha, ztmp2, ok, PP, Ptmp, B_lll,
		     &time_4_lll, &time_4_Hensel, Sol, h);
      
      
      if (verbose >= 3)
        {
          if (success >= 0)
            {
              printf ("Success, %d candidate(s) found\n", Nb);
              if (Nb > 0)
                gmp_printf ("Res[0] is %Zd\n", Res[0]);
            }
          else
            printf("Failure.\n"); 
        }
      
      if (success == -1) {
	/* in case of failure, split the interval under consideration */ 
	failures ++;
	
#ifdef AUTOMATIC      
	/* If too many failures in proportion, reduce ??? */ 
	if ((good > 0) && ((100.0 * failures) / good > FAILURE_UPPERBOUND)){
	  mpz_mul_ui (initial_T, initial_T, 9);
	  mpz_fdiv_q_ui (initial_T, initial_T, 10);
	  if (mpz_sgn (initial_T)==0)
	    mpz_set_ui (initial_T, 1);

#ifndef SINBIG	
	  mpfr_set_d (Mprime, 0.9, MPFR_RNDU);
	  mpfr_pow_ui (Mprime, Mprime, d+1, MPFR_RNDU);
	  mpfr_mul (initial_eps, initial_eps, Mprime, MPFR_RNDU);
	  
	  mpfr_ui_div (Mprime, 1, M, MPFR_RNDN);
	  mpfr_add (Mprime, Mprime, initial_eps, MPFR_RNDN);
	  mpfr_ui_div (Mprime, 1, Mprime, MPFR_RNDN);
	  mpfr_div_ui (Mprime, Mprime, 2, MPFR_RNDN);
	  mpfr_get_z (initial_C, Mprime, MPFR_RNDD);
#endif
	}
#endif

#ifdef SINBIG
      halve_T:
#endif
	mpz_div_ui (T, T, 2);
	
	if (mpz_cmp_ui (T, 0)==0)  {
#ifndef NO_OVERPASS_FAILURES
	  mpz_set_ui (ztmp, 0);
	  number += Check (&ztmp, 1, t0current, t1, Nin, NN, 
			   oneoverM, T, e_in, n, q, yy);
	  total_T += 1.0;  
	  good ++;
#ifdef SINBIG
	  mpz_add (t0current, t0current, q);
#else
	  mpz_add_ui (t0current, t0current, 1);
#endif
	  
	  mpz_set_ui (T, 1);
	  
#else
	  fprintf (stderr, "FAILURE at x =");
	  mpz_out_str (stderr, 10, t0current);
	  fprintf (stderr, "\n");
	  exit (1);
#endif
	}

	mpfr_div_2ui (eps, eps, d+1, MPFR_RNDU);
	mpfr_ui_div (Mprime, 1, M, MPFR_RNDN);
	mpfr_add (Mprime, Mprime, eps, MPFR_RNDN);
	mpfr_ui_div (Mprime, 1, Mprime, MPFR_RNDN);
	mpfr_div_ui (Mprime, Mprime, 2, MPFR_RNDN);
	mpfr_get_z (C, Mprime, MPFR_RNDD);
	mpz_pow_ui (Calpha, C, alpha);
      }
      else {
	/* update total_T */
	total_T += (2.0*(mpz_get_d (T))+1.0);          
	good++;
	
#ifdef AUTOMATIC
	if ((good > 0) && ((100.0 * failures) / good < FAILURE_LOWERBOUND)){
	  mpz_mul_ui (initial_T, initial_T, 10);
	  mpz_fdiv_q_ui (initial_T, initial_T, 9);
	  
	  mpfr_set_d (Mprime, 10.0/9.0, MPFR_RNDU);
	  mpfr_pow_ui (Mprime, Mprime, d+1, MPFR_RNDU);
	  mpfr_mul (initial_eps, initial_eps, Mprime, MPFR_RNDU);
	  
	  mpfr_ui_div (Mprime, 1, M, MPFR_RNDN);
	  mpfr_add (Mprime, Mprime, initial_eps, MPFR_RNDN);
	  mpfr_ui_div (Mprime, 1, Mprime, MPFR_RNDN);
	  mpfr_div_ui (Mprime, Mprime, 2, MPFR_RNDN);
	  mpfr_get_z (initial_C, Mprime, MPFR_RNDD);
	}
#endif
#if DEBUG >= 2
	printf ("eps = ");
	mpfr_out_str (stdout, 10, 10, initial_eps, MPFR_RNDN);
	printf ("\n");
	printf ("total_T is %E\n", total_T);
	printf ("failures is %lu\n", failures);
#endif

	Nb = remove_doubles(Res, Nb);
#ifdef EVAL_CHECK
	int cptr_check, st_check = cputime(); 
	
	for (cptr_check = 0; cptr_check < NBIT_CHECK; cptr_check++)
#endif
	  /* check potential solutions and print solutions */
	  number += Check (Res, Nb, tm, t1, Nin, NN, oneoverM, T, e_in, n, q, yy);
#ifdef EVAL_CHECK
	printf("Check %d\n", (cputime() - st_check) / NBIT_CHECK); 
#endif
	for (i = 0; i < Nb; i++) mpz_clear (Res[i]);
	if (Nb > 0) free(Res); /* ?? */ 
	
#ifdef SINBIG
	mpz_mul (t0current, q, T);
	mpz_add (t0current, t0current, tm);
	mpz_add (t0current, t0current, q);
#else
	mpz_add (t0current, tm, T);
	mpz_add_ui (t0current, t0current, 1);
#endif
	
	/* If T is smaller than initial_T, then reset it */
	mpz_set (T, initial_T);
	mpfr_set (eps, initial_eps, MPFR_RNDU);
	mpz_set (C, initial_C);
	mpz_pow_ui (Calpha, C, alpha);
      }
  }

  hensel_clear (h);

#pragma omp atomic update
  Failures += failures;
#pragma omp atomic update
  Good += good;
#pragma omp atomic update
  Number += number;
#pragma omp critical
  mpz_add (AverageT, AverageT, initial_T);

#ifdef VBSE
#ifdef EVAL_HENSEL
#ifdef EVAL_LLL
  fprintf (stderr, "LLL: %ld ms, Hensel: %ld ms\n",
	   time_4_lll/1000, time_4_Hensel/1000);
#endif
#endif
#endif

  /* Clear everything before leaving */ 
  for (i=0; i<=d; i++) mpfr_clear (P[i]); 
  free (P);

  mpz_clear (t0current);
  mpz_clear (done);

  mpz_clear (C); mpz_clear (initial_C); mpz_clear (initial_T); mpz_clear (T); 
  mpz_clear (tm); mpz_clear (ztmp); mpz_clear (ztmp2); mpz_clear (Calpha); 
  mpz_clear (NZ);

  mpfr_clear (Nin);
  mpfr_clear (tmp);
  mpfr_clear (tmp2);

  mpfr_clear (eps); mpfr_clear (initial_eps); mpfr_clear (NN);  
  mpfr_clear (M); mpfr_clear (Mprime);  mpfr_clear (oneoverM);
  mpfr_clear (yy);
  
  clear_matrix (&W, npols, ncoefs);

  free_space_heuristic (&B_lll, npols, ncoefs);
  free (ok);

  for (j = 0; j < nb; j++)
    {
      mpz_clear (Sol[0][j]);
      mpz_clear	(Sol[1][j]);
    }
  free (Sol[0]);
  free (Sol[1]);

  clear_vec (PP, d+1);  clear_vec (Ptmp, ncoefs);
}

int
main (int argc, char *argv[])
{
  mp_prec_t     prec;
  int           d, alpha, e_in;
  double        n, nn, m, t;
  mpz_t         t0, t1, q;
  mpfr_t        eps, left, right;
  unsigned long nthreads = 1; /* number of threads */
  unsigned long nbr;          /* number of loops for SINBIG */
  unsigned long i;

  /***  BEGINNING OF Initialization and argument parsing ***/ 

  mpz_init (t0);
  mpz_init (t1);
  mpz_init (q);
  mpz_init (AverageT);
  analyse_argv (argc, argv, &near, &prec, &n, &nn, &m, &t, &d, &alpha, &e_in, 
	        t0, t1, &nbr, q, &nthreads);

  if (verbose > 0)
    {
      fprintf (stderr, "Welcome to BaCSeL version %s, using %lu thread(s)\n",
               VERSION, nthreads);
      fprintf (stderr, "Compilation flags: -DBASIS=%d", BASIS);
#ifdef EXP
      fprintf (stderr, " -DEXP");
#endif
#ifdef SIN
      fprintf (stderr, " -DSIN");
#endif
#ifdef COS
      fprintf (stderr, " -DCOS");
#endif
#ifdef SINBIG
      fprintf (stderr, " -DSINBIG");
#endif
#ifdef TWOTOX
      fprintf (stderr, " -DTWOTOX");
#endif
#ifdef LOG2
      fprintf (stderr, " -DLOG2");
#endif
#ifdef LN
      fprintf (stderr, " -DLN");
#endif
#ifdef ONEOVERX
      fprintf (stderr, " -DONEOVERX");
#endif
#ifdef HALL
      fprintf (stderr, " -DHALL");
#endif
#ifdef SQROOT
      fprintf (stderr, " -DSQROOT");
#endif
#ifdef SQRTINVERT
      fprintf (stderr, " -DSQRTINVERT");
#endif
#ifdef NO_OVERPASS_FAILURES
      fprintf (stderr, " -DNO_OVERPASS_FAILURES");
#endif
#ifdef D3A2
      fprintf (stderr, " -DD3A2");
#endif
#ifdef AUTOMATIC
      fprintf (stderr, " -DAUTOMATIC");
#endif
#ifdef VBSE
      fprintf (stderr, " -DVBSE");
#endif
#ifdef DEBUG
      fprintf (stderr, " -DDEBUG=%d", DEBUG);
#endif
      fprintf (stderr, "\n");
      fprintf (stderr, "LLL parameters: delta=%.3f eta=%.3f\n", DELTA, ETA);
#ifdef AUTOMATIC
      fprintf (stderr, "Parameters for AUTOMATIC: FAILURE_LOWERBOUND=%.1f FAILURE_UPPERBOUND=%.1f\n", FAILURE_LOWERBOUND, FAILURE_UPPERBOUND);
#endif
    }

  /* compute left and right bounds */
  mpfr_init2 (eps, 53);
  mpfr_init2 (left, 53);
  mpfr_init2 (right, 53);
  mpfr_set_d (eps, n - ((double) e_in), MPFR_RNDN);
  mpfr_ui_pow (eps, BASIS, eps, MPFR_RNDN); /* eps = B^{p - e} */

  mpfr_set_z (left, t0, MPFR_RNDN);
  mpfr_set_z (right, t1, MPFR_RNDN);
  mpfr_div (left, left, eps, MPFR_RNDN);   /* Min. floating point considered */
  mpfr_div (right, right, eps, MPFR_RNDN); /* Max. floating point considered */

  mpfr_init (one_half);
  mpfr_set_d (one_half, 0.5, MPFR_RNDN);
  
  if (verbose >= 2)
    print_args (left, right, m, n, nn, prec, d, alpha, t, e_in, near, q);

#ifdef SINBIG
  /* Compute the elementary step tau. */

  mpfr_init (eps);
  mpfr_init (left);
  mpfr_init (right);
  mpfr_init2 (tau, LARGE_PREC);
  mpfr_init2 (pi, LARGE_PREC);
  mpfr_init2 (twopi, LARGE_PREC);

  mpfr_set_z (tau, q, MPFR_RNDN);
  {
    mpfr_t tmp;
    mpfr_init (tmp);
    mpfr_set_ui (tmp, BASIS, MPFR_RNDN);
    mpfr_set_d (eps, n - ((double) e_in), MPFR_RNDN);
    mpfr_pow (tmp, tmp, eps, MPFR_RNDN); /* tmp = B^{p - e} */
    mpfr_div (tau, tau, tmp, MPFR_RNDN);
    mpfr_clear (tmp);
  }

  mpfr_set_prec (pi, LARGE_PREC);
  mpfr_set_prec (eps, LARGE_PREC);
  mpfr_const_pi (pi, MPFR_RNDN);
  mpfr_mul_2ui (twopi, pi, 1, MPFR_RNDN);
  mpfr_div (eps, tau, pi, MPFR_RNDN);
  mpfr_rint (eps, eps, MPFR_RNDN);
  mpfr_mul (eps, eps, pi, MPFR_RNDN);
  mpfr_sub (tau, tau, eps, MPFR_RNDN);
  
#if DEBUG >= 2
  printf ("tau="); mpfr_out_str (stdout, 10, 10, tau, MPFR_RNDN); printf ("\n");
#endif
#else
  mpz_set_ui (q, 1); 
#endif

  omp_set_num_threads (nthreads);

#ifndef SINBIG
#pragma omp parallel for
  for (i = 0 ; i < nthreads; i++)
    {
      mpz_t ti0, ti1;
      mpz_init (ti0);
      mpz_init (ti1);
      /* ti0 = t0 + i*floor((t1-t0)/nthreads)
         ti1 = t0 + (i+1)*floor((t1-t0)/nthreads) */
      mpz_sub (ti1, t1, t0);
      mpz_fdiv_q_ui (ti1, ti1, nthreads);
      mpz_mul_ui (ti0, ti1, i);
      mpz_add (ti0, t0, ti0);
      if (i < nthreads - 1)
        {
          mpz_mul_ui (ti1, ti1, i + 1);
          mpz_add (ti1, t0, ti1);
        }
      else
        mpz_set (ti1, t1);
      /* warning: the default MPFR precision is local to each thread */
      mpfr_set_default_prec (prec);
      doit (ti0, ti1, n, nn, m, t, d, alpha, e_in, q);
      mpz_clear (ti0);
      mpz_clear (ti1);
    }
#else
  /* for SINBIG, we loop over the number 'nbr' of arithmetic progressions */
#pragma omp parallel for schedule(dynamic,1)
  for (i = 0 ; i < nbr; i++)
    {
      mpz_t ti0;
      mpz_init (ti0);
      mpz_add_ui (ti0, t0, i);
      mpfr_set_default_prec (prec);
      doit (ti0, t1, n, nn, m, t, d, alpha, e_in, q);
      mpz_clear (ti0);
    }
#endif

  if (verbose >= 2)
    {
      fprintf (stderr, "Failure rate: %ld%%.\n", (100 * Failures) / Good);
      fprintf (stderr, "Number of calls to LLL: %ld. \n", Failures + Good);
      fprintf (stderr, "Found %ld solution(s).\n", Number);
      fprintf (stderr, "Done\n");
#ifdef AUTOMATIC
      {
        mpfr_t tmp;
        mpfr_init2 (tmp, 53);
        mpfr_set_z (tmp, AverageT, MPFR_RNDN);
#ifdef SINBIG
        mpfr_div_ui (tmp, tmp, nbr, MPFR_RNDN);
#else
        mpfr_div_ui (tmp, tmp, nthreads, MPFR_RNDN);
#endif
        mpfr_log (tmp, tmp, MPFR_RNDN);
        mpfr_set_d (eps, (double) BASIS, MPFR_RNDN);
        mpfr_log (eps, eps, MPFR_RNDN);
        mpfr_div (tmp, tmp, eps, MPFR_RNDN);
        fprintf (stderr, "Final size of subintervals (log[%d](T)): %E.\n",
                 BASIS, mpfr_get_d (tmp, MPFR_RNDN));
        mpfr_clear (tmp);
      }
#endif
    }

  mpz_clear (t0);
  mpz_clear (t1);

  mpfr_clear (eps);
  mpfr_clear (left);
  mpfr_clear (right);
#ifdef SINBIG
  mpfr_clear (tau);
  mpfr_clear (pi);
  mpfr_clear (twopi);
#endif
  mpfr_clear (one_half);
  mpz_clear (q);
  mpz_clear (AverageT);

  mpfr_free_cache ();
  return 0;
}
