/* FPLLL wrapper.

  Copyright Paul Zimmermann.  

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

*/

#include <fplll.h>
#include "delta_eta.h" /* for DELTA and ETA */
#include "io.h"        /* for MAYBE_UNUSED */

/* inspired from https://github.com/fplll/fplll/blob/master/fplll/main.cpp */
/* B has d rows and n columns */
void
LLLReduce (mpz_t **B, int d, int n, int precision MAYBE_UNUSED)
{
  ZZ_mat<mpz_t> A(d,n);
  /* See page 195 of Serge's Torres PhD thesis (he is using LM_WRAPPER
     and FT_DEFAULT):
   * LM_WRAPPER: Tries to reduce the matrix with a combination of the
     following methods with increasing precision. Then, runs the proved
     version with the default precision. The floatType must be FT_DEFAULT
     and precision must be 0. The result is guaranteed.
   * LM_PROVED: Proved method. Uses integers to compute dot products. The
     result is guaranteed if floatType=FT_DEFAULT/FT_MPFR and precision=0
     (see above).
   * LM_HEURISTIC: Heuristic method. Uses floating-point numbers to compute
     dot products. The result is not guaranteed. It is more efficient that
     the proved version when the coefficients of the input are large (the
     threshold depends on the floating-point type and precision).
   * LM_FAST: Same as LM_HEURISTIC with floatType=FT_DOUBLE, with a
     special trick to handle larger inputs. */
  LLLMethod method = LM_HEURISTIC;
  /* FT_DEFAULT: LM_WRAPPER: yes
                 LM_PROVED: yes, same as FT_MPFR
		 LM_HEURISTIC: yes, same as FT_DPE
		 LM_FAST: yes, same as FT_DOUBLE
     FT_DOUBLE:  LM_WRAPPER: no
                 LM_PROVED: yes
		 LM_HEURISTIC: yes
		 LM_FAST: yes
     FT_DPE:     LM_WRAPPER: no
                 LM_PROVED: yes
		 LM_HEURISTIC: yes
		 LM_FAST: no
     FT_MPFR:    LM_WRAPPER: no
                 LM_PROVED: yes
		 LM_HEURISTIC: yes
		 LM_FAST: no */
  FloatType float_type = FT_DOUBLE; /* FT_MPFR, FT_DPE, FT_DD, FT_QD, FT_DOUBLE, FT_LONG_DOUBLE */
  int status MAYBE_UNUSED;
  int flags = 0;
  /* copy B into A */
  for (int i = 0; i < d; i++)
    for (int j = 0; j < n; j++)
      A[i][j] = B[i][j];
  /* FIXME: to avoid copy, maybe use the .data() method to access array of
     elements, since rows are vector<Z_NR<mpz_t>>.
     See http://www.cplusplus.com/reference/vector/vector/data/. */
  /* reduce A */
  if (float_type == FT_MPFR)
    status = lll_reduction (A, DELTA, ETA, method, float_type, precision, flags);
  else
    status = lll_reduction (A, DELTA, ETA, method, float_type, flags);
  assert (status == RED_SUCCESS);
  /* copy back A into B */
  for (int i = 0; i < d; i++)
    for (int j = 0; j < n; j++)
      A[i][j].get_mpz(B[i][j]);
  }
